@@ -9,15 +9,23 @@
 #    LICENSE file in the root directory of this source tree.
 ############################################################################
 
-import itertools
+import functools
+import warnings
 
 import numpy as np
 import qutip as qt
 
-from scqubits.core.data_containers import SpectrumData
+import scqubits.settings as settings
+from scqubits.core.central_dispatch import (DispatchClient,
+                                            CENTRAL_DISPATCH)
+from scqubits.core.descriptors import ReadOnlyProperty, WatchedProperty
 from scqubits.core.harmonic_osc import Oscillator
-from scqubits.settings import IN_IPYTHON, TQDM_KWARGS
-from scqubits.utils.spectrum_utils import get_matrixelement_table, convert_esys_to_ndarray
+from scqubits.core.spec_lookup import SpectrumLookup
+from scqubits.core.storage import SpectrumData
+from scqubits.settings import IN_IPYTHON
+from scqubits.utils.misc import InfoBar
+from scqubits.utils.processing_switch import get_map_method
+from scqubits.utils.spectrum_utils import convert_operator_to_qobj, recast_esys_mapdata
 
 if IN_IPYTHON:
     from tqdm.notebook import tqdm
@@ -25,12 +33,13 @@ else:
     from tqdm import tqdm
 
 
-class InteractionTerm:
+class InteractionTerm(DispatchClient):
     """
     Class for specifying a term in the interaction Hamiltonian of a composite Hilbert space, and constructing
-    the Hamiltonian in qutip.Qobj format. The expected form of the interaction term is g A B, where g is the
-    interaction strength, A an operator in subsystem 1 and B and operator in subsystem 2.
-
+    the Hamiltonian in qutip.Qobj format. The expected form of the interaction term is of two possible types:
+    1. V = g A B, where A, B are Hermitean operators in two specified subsystems,
+    2. V = g A B + h.c., where A, B may be non-Hermitean
+    
     Parameters
     ----------
     g_strength: float
@@ -41,51 +50,57 @@ class InteractionTerm:
         the two subsystems involved in the interaction
     op1, op2: str or ndarray
         names of operators in the two subsystems
-    evecs1, evecs2: ndarray
-        bare eigenvectors allowing the calculation of op1, op2 in the two bare eigenbases
+    add_hc: bool, optional (default=False)
+        If set to True, the interaction Hamiltonian is of type 2, and the Hermitean conjugate is added.
     """
-    def __init__(self, g_strength, hilbertspace, subsys1, op1, subsys2, op2, evecs1=None, evecs2=None):
+    g_strength = WatchedProperty('INTERACTIONTERM_UPDATE')
+    subsys1 = WatchedProperty('INTERACTIONTERM_UPDATE')
+    subsys2 = WatchedProperty('INTERACTIONTERM_UPDATE')
+    op1 = WatchedProperty('INTERACTIONTERM_UPDATE')
+    op2 = WatchedProperty('INTERACTIONTERM_UPDATE')
+
+    def __init__(self, g_strength, subsys1, op1, subsys2, op2, add_hc=False, hilbertspace=None):
+        if hilbertspace:
+            warnings.warn("`hilbertspace` is no longer a parameter for initializing an InteractionTerm object.",
+                          FutureWarning)
         self.g_strength = g_strength
-        self.hilbertspace = hilbertspace
         self.subsys1 = subsys1
         self.op1 = op1
         self.subsys2 = subsys2
         self.op2 = op2
-        self.evecs1 = evecs1
-        self.evecs2 = evecs2
-
-    def hamiltonian(self, evecs1=None, evecs2=None):
-        """
-        Parameters
-        ----------
-        evecs1, evecs2: ndarray, optional
-            subsystem eigenvectors used to calculated interaction Hamiltonian; calculated on the fly if not given
-
-        Returns
-        -------
-        qutip.Qobj operator
-            interaction Hamiltonian
-        """
-        interaction_op1 = self.hilbertspace.identity_wrap(self.op1, self.subsys1, op_in_eigenbasis=False, evecs=evecs1)
-        interaction_op2 = self.hilbertspace.identity_wrap(self.op2, self.subsys2, op_in_eigenbasis=False, evecs=evecs2)
-        return self.g_strength * interaction_op1 * interaction_op2
+        self.add_hc = add_hc
 
 
-class HilbertSpace(list):
+class HilbertSpace(DispatchClient):
     """Class holding information about the full Hilbert space, usually composed of multiple subsystems.
     The class provides methods to turn subsystem operators into operators acting on the full Hilbert space, and
     establishes the interface to qutip. Returned operators are of the `qutip.Qobj` type. The class also provides methods
     for obtaining eigenvalues, absorption and emission spectra as a function of an external parameter.
     """
+    osc_subsys_list = ReadOnlyProperty()
+    qbt_subsys_list = ReadOnlyProperty()
+    lookup = ReadOnlyProperty()
+    interaction_list = WatchedProperty('INTERACTIONLIST_UPDATE')
 
     def __init__(self, subsystem_list, interaction_list=None):
-        list.__init__(self, subsystem_list)
-        self.interaction_list = interaction_list
-        self.state_lookup_table = None
-        self.osc_subsys_list = [(index, subsys) for (index, subsys) in enumerate(self)
-                                if isinstance(subsys, Oscillator)]
-        self.qbt_subsys_list = [(index, subsys) for (index, subsys) in enumerate(self)
-                                if not isinstance(subsys, Oscillator)]
+        self._subsystems = tuple(subsystem_list)
+        if interaction_list:
+            self.interaction_list = tuple(interaction_list)
+        else:
+            self.interaction_list = None
+
+        self._lookup = None
+        self._osc_subsys_list = [(index, subsys) for (index, subsys) in enumerate(self)
+                                 if isinstance(subsys, Oscillator)]
+        self._qbt_subsys_list = [(index, subsys) for (index, subsys) in enumerate(self)
+                                 if not isinstance(subsys, Oscillator)]
+
+        CENTRAL_DISPATCH.register('QUANTUMSYSTEM_UPDATE', self)
+        CENTRAL_DISPATCH.register('INTERACTIONTERM_UPDATE', self)
+        CENTRAL_DISPATCH.register('INTERACTIONLIST_UPDATE', self)
+
+    def __getitem__(self, index):
+        return self._subsystems[index]
 
     def __str__(self):
         output = '====== HilbertSpace object ======\n'
@@ -93,6 +108,34 @@ class HilbertSpace(list):
             output += '\n' + str(subsystem) + '\n'
         return output
 
+    def index(self, item):
+        return self._subsystems.index(item)
+
+    def _get_metadata_dict(self):
+        meta_dict = {}
+        for index, subsystem in enumerate(self):
+            subsys_meta = subsystem._get_metadata_dict()
+            renamed_subsys_meta = {}
+            for key in subsys_meta.keys():
+                renamed_subsys_meta[type(subsystem).__name__ + str(index) + '_' + key] = subsys_meta[key]
+            meta_dict.update(renamed_subsys_meta)
+        return meta_dict
+
+    def receive(self, event, sender, **kwargs):
+        if self.lookup is not None:
+            if event == 'QUANTUMSYSTEM_UPDATE' and sender in self:
+                self.broadcast('HILBERTSPACE_UPDATE')
+                self._lookup._out_of_sync = True
+                # print('Lookup table now out of sync')
+            elif event == 'INTERACTIONTERM_UPDATE' and sender in self.interaction_list:
+                self.broadcast('HILBERTSPACE_UPDATE')
+                self._lookup._out_of_sync = True
+                # print('Lookup table now out of sync')
+            elif event == 'INTERACTIONLIST_UPDATE' and sender is self:
+                self.broadcast('HILBERTSPACE_UPDATE')
+                self._lookup._out_of_sync = True
+                # print('Lookup table now out of sync')
+
     @property
     def subsystem_dims(self):
         """Returns list of the Hilbert space dimensions of each subsystem
@@ -118,7 +161,51 @@ class HilbertSpace(list):
         Returns
         -------
         int"""
-        return len(self)
+        return len(self._subsystems)
+
+    def generate_lookup(self):
+        bare_specdata_list = []
+        for index, subsys in enumerate(self):
+            evals, evecs = subsys.eigensys(evals_count=subsys.truncated_dim)
+            bare_specdata_list.append(SpectrumData(energy_table=[evals], state_table=[evecs],
+                                                   system_params=subsys.__dict__))
+
+        evals, evecs = self.eigensys(evals_count=self.dimension)
+        dressed_specdata = SpectrumData(energy_table=[evals], state_table=[evecs],
+                                        system_params=self._get_metadata_dict())
+        self._lookup = SpectrumLookup(self, bare_specdata_list=bare_specdata_list, dressed_specdata=dressed_specdata)
+
+    def eigenvals(self, evals_count=6):
+        """Calculates eigenvalues of the full Hamiltonian using `qutip.Qob.eigenenergies()`.
+
+        Parameters
+        ----------
+        evals_count: int, optional
+            number of desired eigenvalues/eigenstates
+
+        Returns
+        -------
+        eigenvalues: ndarray of float
+        """
+        hamiltonian_mat = self.hamiltonian()
+        return hamiltonian_mat.eigenenergies(eigvals=evals_count)
+
+    def eigensys(self, evals_count):
+        """Calculates eigenvalues and eigenvectore of the full Hamiltonian using `qutip.Qob.eigenstates()`.
+
+        Parameters
+        ----------
+        evals_count: int, optional
+            number of desired eigenvalues/eigenstates
+
+        Returns
+        -------
+        evals: ndarray of float
+        evecs: ndarray of Qobj kets
+        """
+        hamiltonian_mat = self.hamiltonian()
+        evals, evecs = hamiltonian_mat.eigenstates(eigvals=evals_count)
+        return evals, evecs
 
     def diag_operator(self, diag_elements, subsystem):
         """For given diagonal elements of a diagonal operator in `subsystem`, return the `Qobj` operator for the
@@ -182,26 +269,7 @@ class HilbertSpace(list):
         -------
         qutip.Qobj operator
         """
-        dim = subsystem.truncated_dim
-
-        if isinstance(operator, np.ndarray):
-            if op_in_eigenbasis is False:
-                if evecs is None:
-                    _, evecs = subsystem.eigensys(evals_count=subsystem.truncated_dim)
-                operator_matrixelements = get_matrixelement_table(operator, evecs)
-                subsys_operator = qt.Qobj(inpt=operator_matrixelements)
-            else:
-                subsys_operator = qt.Qobj(inpt=operator[:dim, :dim])
-        elif isinstance(operator, str):
-            if evecs is None:
-                _, evecs = subsystem.eigensys(evals_count=subsystem.truncated_dim)
-            operator_matrixelements = subsystem.matrixelement_table(operator, evecs=evecs)
-            subsys_operator = qt.Qobj(inpt=operator_matrixelements)
-        elif isinstance(operator, qt.Qobj):
-            subsys_operator = operator
-        else:
-            raise TypeError('Unsupported operator type: ', type(operator))
-
+        subsys_operator = convert_operator_to_qobj(operator, subsystem, op_in_eigenbasis, evecs)
         operator_identitywrap_list = [qt.operators.qeye(the_subsys.truncated_dim) for the_subsys in self]
         subsystem_index = self.get_subsys_index(subsystem)
         operator_identitywrap_list[subsystem_index] = subsys_operator
@@ -255,7 +323,7 @@ class HilbertSpace(list):
         """
         return self.index(subsys)
 
-    def get_bare_hamiltonian(self):
+    def bare_hamiltonian(self):
         """
         Returns
         -------
@@ -268,7 +336,12 @@ class HilbertSpace(list):
             bare_hamiltonian += self.diag_hamiltonian(subsys, evals)
         return bare_hamiltonian
 
-    def get_hamiltonian(self):
+    def get_bare_hamiltonian(self):
+        """Deprecated, use `bare_hamiltonian()` instead."""
+        warnings.warn('bare_hamiltonian() is deprecated, use bare_hamiltonian() instead', FutureWarning)
+        return self.bare_hamiltonian()
+
+    def hamiltonian(self):
         """
 
         Returns
@@ -276,13 +349,43 @@ class HilbertSpace(list):
         qutip.qobj
             Hamiltonian of the composite system, including the interaction between components
         """
-        hamiltonian = self.get_bare_hamiltonian()
-        for interaction_term in self.interaction_list:
-            hamiltonian += interaction_term.hamiltonian()
+        return self.bare_hamiltonian() + self.interaction_hamiltonian()
+
+    def get_hamiltonian(self):
+        """Deprecated, use `hamiltonian()` instead."""
+        return self.hamiltonian()
+
+    def interaction_hamiltonian(self):
+        """
+        Returns
+        -------
+        qutip.Qobj operator
+            interaction Hamiltonian
+        """
+        if self.interaction_list is None:
+            return 0
+
+        hamiltonian = [self.interactionterm_hamiltonian(term) for term in self.interaction_list]
+        return sum(hamiltonian)
+
+    def interactionterm_hamiltonian(self, interactionterm, evecs1=None, evecs2=None):
+        interaction_op1 = self.identity_wrap(interactionterm.op1, interactionterm.subsys1, evecs=evecs1)
+        interaction_op2 = self.identity_wrap(interactionterm.op2, interactionterm.subsys2, evecs=evecs2)
+        hamiltonian = interactionterm.g_strength * interaction_op1 * interaction_op2
+        if interactionterm.add_hc:
+            return hamiltonian + hamiltonian.conj()
         return hamiltonian
 
-    def get_spectrum_vs_paramvals(self, hamiltonian_func, param_vals, evals_count=10, get_eigenstates=False,
-                                  param_name="external_parameter", filename=None):
+    def _esys_for_paramval(self, paramval, update_hilbertspace, evals_count):
+        update_hilbertspace(paramval)
+        return self.eigensys(evals_count)
+
+    def _evals_for_paramval(self, paramval, update_hilbertspace, evals_count):
+        update_hilbertspace(paramval)
+        return self.eigenvals(evals_count)
+
+    def get_spectrum_vs_paramvals(self, param_vals, update_hilbertspace, evals_count=10, get_eigenstates=False,
+                                  param_name="external_parameter", num_cpus=settings.NUM_CPUS):
         """Return eigenvalues (and optionally eigenstates) of the full Hamiltonian as a function of a parameter.
         Parameter values are specified as a list or array in `param_vals`. The Hamiltonian `hamiltonian_func`
         must be a function of that particular parameter, and is expected to internally set subsystem parameters.
@@ -290,127 +393,40 @@ class HilbertSpace(list):
 
         Parameters
         ----------
-        hamiltonian_func: function of one parameter
-            function returning the Hamiltonian in `qutip.Qobj` format
         param_vals: ndarray of floats
             array of parameter values
+        update_hilbertspace: function
+            update_hilbertspace(param_val) specifies how a change in the external parameter affects
+            the Hilbert space components
         evals_count: int, optional
             number of desired energy levels (default value = 10)
         get_eigenstates: bool, optional
             set to true if eigenstates should be returned as well (default value = False)
         param_name: str, optional
             name for the parameter that is varied in `param_vals` (default value = "external_parameter")
-        filename: str, optional
-            write data to file if path/filename is provided (default value = None)
+        num_cpus: int, optional
+            number of cores to be used for computation (default value: settings.NUM_CPUS)
 
         Returns
         -------
         SpectrumData object
         """
-        paramvals_count = len(param_vals)
-
-        eigenenergy_table = np.empty((paramvals_count, evals_count))
+        target_map = get_map_method(num_cpus)
         if get_eigenstates:
-            eigenstatesQobj_table = [0] * paramvals_count
+            func = functools.partial(self._esys_for_paramval, update_hilbertspace=update_hilbertspace,
+                                     evals_count=evals_count)
+            with InfoBar("Parallel computation of eigenvalues [num_cpus={}]".format(num_cpus), num_cpus):
+                eigensystem_mapdata = list(target_map(func, tqdm(param_vals, desc='Spectral data', leave=False,
+                                                                 disable=(num_cpus > 1))))
+            eigenvalue_table, eigenstate_table = recast_esys_mapdata(eigensystem_mapdata)
         else:
-            eigenstatesQobj_table = None
+            func = functools.partial(self._evals_for_paramval,  update_hilbertspace=update_hilbertspace,
+                                     evals_count=evals_count)
+            with InfoBar("Parallel computation of eigensystems [num_cpus={}]".format(num_cpus), num_cpus):
+                eigenvalue_table = list(target_map(func, tqdm(param_vals, desc='Spectral data', leave=False,
+                                                              disable=(num_cpus > 1))))
+            eigenvalue_table = np.asarray(eigenvalue_table)
+            eigenstate_table = None
 
-        for param_index, paramval in tqdm(enumerate(param_vals), total=len(param_vals), **TQDM_KWARGS):
-            paramval = param_vals[param_index]
-            hamiltonian = hamiltonian_func(paramval)
-
-            if get_eigenstates:
-                eigenenergies, eigenstates_Qobj = hamiltonian.eigenstates(eigvals=evals_count)
-                eigenenergy_table[param_index] = eigenenergies
-                eigenstatesQobj_table[param_index] = eigenstates_Qobj
-            else:
-                eigenenergies = hamiltonian.eigenenergies(eigvals=evals_count)
-                eigenenergy_table[param_index] = eigenenergies
-
-        spectrumdata = SpectrumData(param_name, param_vals, eigenenergy_table, self.__dict__,
-                                    state_table=eigenstatesQobj_table)
-        if filename:
-            spectrumdata.filewrite(filename)
-
-        return spectrumdata
-
-    def generate_state_lookup_table(self, spectrum_data, param_index=0):
-        """
-        Create a lookup table that associates each dressed-state index with the corresponding bare-state product state
-        index (whenever possible). Usually to be saved as self.state_lookup_table.
-
-        Parameters
-        ----------
-        spectrum_data: SpectrumData
-        param_index: int
-            indices > 0 become relevant when using ParameterSweep
-
-        Returns
-        -------
-        list(int), list(tuple)
-            dressed indices, corresponding bare indices
-        """
-        dims = self.subsystem_dims
-        product_dim = np.prod(dims)
-        overlap_matrix = convert_esys_to_ndarray(spectrum_data.state_table[param_index])
-        dressed_indices = []
-        for bare_basis_index in range(product_dim):
-            max_position = (np.abs(overlap_matrix[:, bare_basis_index])).argmax()
-            max_overlap = np.abs(overlap_matrix[max_position, bare_basis_index])
-            if max_overlap < 0.5:
-                dressed_indices.append(None)
-            else:
-                dressed_indices.append(max_position)
-        basis_label_ranges = [list(range(dims[subsys_index])) for subsys_index in range(self.subsystem_count)]
-        basis_labels_list = list(itertools.product(*basis_label_ranges))
-        return [dressed_indices, basis_labels_list]
-
-    def lookup_dressed_index(self, bare_labels, param_index=0):
-        """
-        Parameters
-        ----------
-        bare_labels: tuple of ints
-            bare_labels = (index, index2, ...)
-        param_index: int
-            index of parameter value of interest
-
-        Returns
-        -------
-        int
-            dressed state index closest to the specified bare state
-        """
-        try:
-            lookup_position = self.state_lookup_table[param_index][1].index(bare_labels)
-        except ValueError:
-            return None
-        return self.state_lookup_table[param_index][0][lookup_position]
-
-    def lookup_bare_index(self, dressed_index, param_index=0):
-        """
-        For given dressed index, look up the corresponding bare index from the state_lookup_table.
-
-        Parameters
-        ----------
-        dressed_index: int
-        param_index: int
-
-        Returns
-        -------
-        tuple(int)
-        """
-        try:
-            lookup_position = self.state_lookup_table[param_index][0].index(dressed_index)
-        except ValueError:
-            return None
-        basis_labels = self.state_lookup_table[param_index][1][lookup_position]
-        return basis_labels
-
-    def _get_metadata_dict(self):
-        meta_dict = {}
-        for index, subsystem in enumerate(self):
-            subsys_meta = subsystem._get_metadata_dict()
-            renamed_subsys_meta = {}
-            for key in subsys_meta.keys():
-                renamed_subsys_meta[type(subsystem).__name__ + str(index) + '_' + key] = subsys_meta[key]
-            meta_dict.update(renamed_subsys_meta)
-        return meta_dict
+        return SpectrumData(eigenvalue_table, self._get_metadata_dict(), param_name, param_vals,
+                            state_table=eigenstate_table)
